/* $Id: divactrl.c,v 1.5 2002/10/21 08:15:27 armin Exp $
 *
 * DIVA-ISDN driver for Linux. (Control-Utility)
 *
 * Copyright 2001-2004 Cytronics & Melware (info@melware.de)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Log: divactrl.c,v $
 * Revision 1.5  2002/10/21 08:15:27  armin
 * update to new version, new interface for driver 2.0,
 * big-endian fixes, AIX compatibility.
 *
 * Revision 1.4  2002/07/22 15:19:22  armin
 * Added some configuration tools.
 *
 * Revision 1.3  2002/07/19 09:18:27  armin
 * Now compilable without kernel sources using its own copy of kernel-
 * include files. Fixed cross-compile/different-target rules.
 * Added check for kernel driver version.
 *
 * Revision 1.2  2002/07/18 14:43:00  armin
 * New version 2.0 with latest firmware and ditrace debug tool.
 *
 * Revision 1.1  2001/10/12 09:50:24  armin
 * Initial checkin of new version v1.2
 *
 *
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <malloc.h>
#include <ctype.h>

#include <platform.h>
#include <di_defs.h>
#include <cardtype.h>
#include <divasync.h>
#undef OK
#include <pc.h>
#include <di.h>

extern int check_device_nodes(void);
extern void fixup_card_number(char *);
extern int divaload_main (int, char**);
extern int mantool_main (int, char**);
extern int mlog_main (int, char**);
extern int dchannel_main (int, char**);
extern int ditrace_main (int, char**);
#ifdef AIX
extern int aixmodule_main (int, char**);
#endif

extern word xlog(FILE *stream, void * buffer);

void no_printf (char * x ,...)
{
  /* dummy debug function */
}
DIVA_DI_PRINTF my_dprintf = no_printf;

char *cmd;

void divactrl_usage(void)
{
	fprintf(stderr,"divactrl utility version 2.1.1                    (c) 2004 Cytronics & Melware\n");
	fprintf(stderr,"BUILD (%s-%s-%s)\n", DIVA_BUILD, __DATE__, __TIME__ );
	fprintf(stderr,"usage: %s load      [arg1 arg2 argn]  (load and start card)\n", cmd);
	fprintf(stderr,"       %s ctrl      [arg1 arg2 argn]  (card controlling functions) \n", cmd);
	fprintf(stderr,"       %s mantool   [arg1 arg2 argn]  (B-Channel state trace utility)\n", cmd);
	fprintf(stderr,"       %s mlog      [arg1 arg2 argn]  (trace/log utility)\n", cmd);
	fprintf(stderr,"       %s dchannel  [arg1 arg2 argn]  (D-Channel log)\n", cmd);
	fprintf(stderr,"       %s ditrace   [arg1 arg2 argn]  (Driver log tool)\n", cmd);
#ifdef AIX
	fprintf(stderr,"       %s aixmodule [arg1 arg2 argn]  (load/config kernel-driver tool)\n", cmd);
#endif
	fprintf(stderr,"       %s config                      (start configuration tool)\n", cmd);
	fprintf(stderr,"\n");
	fprintf(stderr,"Type %s [load|ctrl|mantool|mlog|ditrace] -h for function related help.\n", cmd);

	exit(-1);
}

/*
** The main !
*/
int main(int argc, char **argv)
{
  char *newargv[100];
  int newarg;
  int x;
  int ret = 0;

  cmd = argv[0];

  if(argc < 2)
    divactrl_usage();
  if (!strcmp(argv[1], "-h"))
    divactrl_usage();
  if (!strcmp(argv[1], "--help"))
    divactrl_usage();

  /*
    start the config tool/script
   */
  if(!strcmp(argv[1], "config")) {
	char *argv[] = { "sh", "-c", DATADIR"/Config", 0 };
	execve("/bin/sh", argv, NULL);
	exit(0);
  }

  /*
    check the device nodes
   */
  check_device_nodes();

  
  /*
    build new arg(v)
  */
  newarg = 1;
  newargv[0] = malloc(160);
  for (x = 2; x < argc; x++)
  {
    newargv[newarg] = malloc(strlen(argv[x]) + 1);
    strcpy(newargv[newarg++],argv[x]);
  } 
  newargv[newarg] = NULL;

  /*
    fixup card number specified with -c
  */
  for (x = 1; x < newarg; x++) {
	  if ((!strcmp(newargv[x], "-c")) && (x < (newarg - 1))) {
		  fixup_card_number(newargv[x+1]);
		  break;
	  }
  }

  /*
    divaload
  */
  if(!strcmp(argv[1], "load")) {
    strcpy(newargv[0], "divactrl load");
    ret = divaload_main(newarg, newargv);
    if(ret){
      printf("In case of error, you can try option -Debug in additon.\n");
      printf("Type 'divactrl load -h' for help.\n");
	}
    return(ret);
  }
  if(!strcmp(argv[1], "ctrl")) {
    strcpy(newargv[0], "divactrl ctrl");
    ret = divaload_main(newarg, newargv);
    if(ret){
      printf("In case of error, you can try option -Debug in additon.\n");
      printf("Type 'divactrl ctrl -h' for help.\n");
	}
    return(ret);
  }

  /*
    mantool
  */
  if(!strcmp(argv[1], "mantool")) {
    strcpy(newargv[0], "divactrl mantool");
    ret = mantool_main(newarg, newargv);
    if(ret)
      printf("Type 'divactrl mantool -h' for help.\n");
    return(ret);
  }
 
	/*
		ditrace
	*/
	if(!strcmp(argv[1], "ditrace")) {
			strcpy(newargv[0], "divactrl ditrace");
			ret = ditrace_main(newarg, newargv);
			if(ret)
					printf("Type 'divactrl ditrace -h' for help.\n");
			return(ret);
	}
	  
  /*
    mlog
  */
  if(!strcmp(argv[1], "mlog")) {
    strcpy(newargv[0], "divactrl mlog");
    ret = mlog_main(newarg, newargv);
    if(ret)
      printf("Type 'divactrl mlog -h' for help.\n");
    return(ret);
  }
 
  /*
    dchannel
  */
  if(!strcmp(argv[1], "dchannel")) {
    strcpy(newargv[0], "divactrl dchannel");
    ret = dchannel_main(newarg, newargv);
    if(ret)
      printf("Type 'divactrl dchannel -h' for help.\n");
    return(ret);
  }
 
#ifdef AIX
  if(!strcmp(argv[1], "aixmodule")) {
    strcpy(newargv[0], "divactrl aixmodule");
    ret = aixmodule_main(newarg, newargv);
    return(ret);
  }
#endif

  divactrl_usage();
  return(0);
}





/*
**  This section does contain all functions that are necessary to
**  resolve all externals unused in the user mode configurator
*/
void
diva_os_prepare_pri_functions (PISDN_ADAPTER IoAdapter)
{
}

byte
scom_test_int(ADAPTER * a)
{
        printf ("test int ?\n");
        exit (10);
        return (0);
}

byte
pr_dpc(ADAPTER * a)
{
        printf ("pr_dpc ?\n");
        exit (10);
        return (0);
}

void
pr_out(ADAPTER * a)
{
        printf ("pr_out ?\n");
        exit (10);
}

void
diva_os_prepare_pri2_functions (PISDN_ADAPTER IoAdapter)
{
}

void
pr_stream (ADAPTER * a)
{
}

int
diva_os_schedule_soft_isr (diva_os_soft_isr_t* psoft_isr)
{
        return (0);
}

void
scom_clear_int(ADAPTER * a)
{
        printf ("scom_clear_int ?\n");
        exit (10);
}

void
diva_os_wait (dword wait_mS)
{
}

void
diva_os_sleep (dword wait_mS)
{
}

long
diva_os_atomic_increment (diva_os_atomic_t* p)
{
  printf ("atomic increment ?\n");
  exit (10);
  return (0);
}

long
diva_os_atomic_decrement (diva_os_atomic_t *p)
{
  printf ("atomic decrement ?\n");
  exit (10);
  return (0); 
}

void
diva_xdi_provide_istream_info (ADAPTER* a,
                               diva_xdi_stream_interface_t* pi)
{
}

void
inppw_buffer (void* addr, void* p, int lengrh)
{
}

void
diva_os_enter_spin_lock (diva_os_spin_lock_t* pspin,
                         diva_os_spin_lock_magic_t* pmagic,
                         void* debug)
{
}

void
diva_os_leave_spin_lock (diva_os_spin_lock_t* pspin,
                         const diva_os_spin_lock_magic_t* pmagic,
                         void* debug)
{
}

