/* $Id: nodes.c,v 1.4 2003/08/29 12:07:51 schindler Exp $
 *
 * divactrl device node check 
 *
 * (c) 2003 Cytronics & Melware
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>


static unsigned int get_device_major(const char *device)
{
	unsigned int major = 0;
	FILE *fp;
	unsigned char line[1024];
	char *p;

	if ((fp = fopen("/proc/devices", "r"))) {
		while(!feof(fp)) {
			line[0] = 0;
			fgets(line, sizeof(line) - 1, fp);
			if (line[strlen(line) - 1] == '\n')
				line[strlen(line) - 1] = 0;
			if ((p = strstr(line, device))) {
				if (strlen(p) == strlen(device)) {
					major = atoi(line);
					break;
				}
			}
		}
		fclose(fp);
	}

	return(major);
}

static int get_node_major(const char *devname)
{
	struct stat st;

	if (stat(devname, &st))
		return(0);

	if ((st.st_mode & S_IFMT) != S_IFCHR)
		return(0);

	return(major(st.st_rdev));
}

static int check_node(const char *name)
{
	unsigned char devname[1024];
	int dmajor, nmajor;

	sprintf(devname, "/dev/%s", name);

	if (!(dmajor = get_device_major(name))) {
		/* printf("%s has no major, removing node.\n", devname); */
		unlink(devname);
		return(0);
	}

	nmajor = get_node_major(devname);

	if (nmajor != dmajor) {
		/* printf("%s has wrong major(%d->%d), changing node.\n", devname, nmajor, dmajor); */
		unlink(devname);
		mknod(devname, S_IFCHR | 0666, makedev(dmajor, 0));
	} else {
		/* printf("%s has correct node major %d.\n", devname, dmajor); */
	}
	
	return(1);	
}

int check_device_nodes(void)
{
	struct stat st;

	/* if devfs is used, node-check is not necessary */
	if (!(stat("/dev/.devfsd", &st))) {
		/* printf("devfs used, not checking nodes.\n"); */
		return(0);
	}
	/* printf("No devfs, checking nodes.\n"); */

	check_node("DivasIDI");
	
	check_node("DivasMAINT");
	
	if (!(check_node("Divas"))) {
		/* fprintf(stderr, "Cannot find divas driver, ERROR.\n"); */
		return(0);
	}

	return(1);
}

