
/*
 *
  Copyright (c) Eicon Networks, 2002.
 *
  This source file is supplied for the use with
  Eicon Networks range of DIVA Server Adapters.
 *
  Eicon File Revision :    2.1
 *
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 *
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY OF ANY KIND WHATSOEVER INCLUDING ANY
  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU General Public License for more details.
 *
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
#include "platform.h"
#include "di_defs.h"
#include "pc.h"
#include "pr_pc.h"
#include "di.h"
#include "mi_pc.h"
#include "pc_maint.h"
#include "divasync.h"
#include "io.h"
#include "dsp_defs.h"
#include "pc_init.h"
#include "helpers.h"
/* --------------------------------------------------------------------------
  Get protocols features from the protocol file for MIPS based cards
  -------------------------------------------------------------------------- */
dword diva_get_protocol_file_features  (byte* File,
                    int offset,
                    char *IdStringBuffer,
                    dword IdBufferSize) {
 byte ch, *features = NULL ;
 dword i, featureMask = 0 ;
 for ( File += offset, i = 0 ;
       (i < IdBufferSize - 1) && (File[i] >= ' ') && (File[i] < 0x7F) ;
       ++i )
 {
  IdStringBuffer[i] = (char)(File[i]) ;
  if ( (File[i] == '[') && (File[i+1] == 'F') && (File[i+2] == '#') )
   features = (byte*)(&IdStringBuffer[i+3]) ;
 }
 IdStringBuffer[i] = '\0' ;
 if ( features )
 {
  while ( (ch = *features++) != 0 )
  {
   if ( (ch >= '0') && (ch <= '9') )
   {
    featureMask = (featureMask << 4) | (ch - '0') ;
    continue ;
   }
   ch &= 0xDF ;
   if ( (ch >= 'A') && (ch <= 'F') )
   {
    featureMask = (featureMask << 4) | (ch - 'A' + 10) ;
    continue ;
   }
   break ;
  }
 }
 return (featureMask) ;
}
/* --------------------------------------------------------------------------
  Local helpers used for protocol configuration
  -------------------------------------------------------------------------- */
static void
CopyNumber (ADAPTER *a, unsigned char *number, unsigned long offset)
{
 unsigned long i ;
 for ( i = 0 ; number[i] ; ++i, ++offset )
 {
  a->ram_out (a, (void *)offset, number[i]) ;
 }
 a->ram_out (a, (void *)offset, 0) ;
}
static unsigned long
CopyUlongIfNonzero (ADAPTER *a, unsigned char type, unsigned long value,
      unsigned char length, unsigned long offset)
{
 unsigned long i;
 if (value)
 {
  a->ram_out (a, (void *)offset, type) ;
  offset++ ;
  if ( !(type & 0x80) )
  {
   a->ram_out (a, (void *)offset, length) ;
   offset++ ;
   for ( i = 0 ; i < length ; i++ )
   {
    a->ram_out (a, (void *)offset, (byte) value) ;
    offset++ ;
    value >>= 8;
   }
  }
 }
 return (offset);
}
/* --------------------------------------------------------------------------
  Write configuration parameters to the card
  -------------------------------------------------------------------------- */
void diva_configure_protocol (PISDN_ADAPTER IoAdapter) {
 ADAPTER *a = &IoAdapter->a ;
 unsigned long offset = 224 ;
 unsigned long xdiFeatures ;
 int i ;
 DBG_REG(("  Protocol = %s (id %d prot %d)",
   IoAdapter->protocol_name, IoAdapter->protocol_id,
    IoAdapter->ProtVersion & ~0x80))
 if ( IoAdapter->tei )
 {
  DBG_REG(("  TEI = %ld", IoAdapter->tei))
 }
 else
 {
  DBG_REG(("  TEI = automatic"))
 }
 DBG_REG(("  NT2 = %ld", IoAdapter->nt2))
 DBG_REG(("  DidLen = %ld", IoAdapter->DidLen))
 if ( IoAdapter->WatchDog )
 {
  DBG_REG(("  Watchdog = %ld", IoAdapter->WatchDog))
 }
 if ( IoAdapter->Permanent )
 {
  DBG_REG(("  Permanent = %ld", IoAdapter->Permanent))
 }
 if ( IoAdapter->StableL2 )
 {
  DBG_REG(("  StableL2 = %ld", IoAdapter->StableL2))
 }
 if ( IoAdapter->NoOrderCheck )
 {
  DBG_REG(("  NoOrderCheck = %ld", IoAdapter->NoOrderCheck))
 }
 if ( IoAdapter->ForceLaw )
 {
  DBG_REG(("  ForceLaw = %ld", IoAdapter->ForceLaw))
 }
 if ( IoAdapter->LowChannel )
 {
  DBG_REG(("  LowChannel = %ld", IoAdapter->LowChannel))
 }
 if ( IoAdapter->NoHscx30 ) {
  DBG_REG(("  NoHscx30   = %ld", IoAdapter->NoHscx30))
 }
 else{
  if(IoAdapter->Properties.Channels > 2 && 
    IoAdapter->Channels <= (IoAdapter->InitialDspInfo >> 16) &&
    IoAdapter->Channels < IoAdapter->Properties.Channels){ 
   IoAdapter->NoHscx30=1;  
   DBG_REG(("  NoHscx30   = %ld", IoAdapter->NoHscx30))
  }
 }
 if ( IoAdapter->Properties.Channels <= 2 )
 {
  DBG_REG(("  OAD1  = \"%s\"", IoAdapter->Oad1))
  DBG_REG(("  OSA1  = \"%s\"", IoAdapter->Osa1))
  DBG_REG(("  SPID1 = \"%s\"", IoAdapter->Spid1))
  DBG_REG(("  OAD2  = \"%s\"", IoAdapter->Oad2))
  DBG_REG(("  OSA2  = \"%s\"", IoAdapter->Osa2))
  DBG_REG(("  SPID2 = \"%s\"", IoAdapter->Spid2))
 }
 else
 {
  DBG_REG(("  CRC4  = %ld", IoAdapter->crc4))
  if ( (IoAdapter->protocol_id == PROTTYPE_RBSCAS) && (i = IoAdapter->Spid1[0]) ) {
   DBG_REG(("  RbsAnswerDelay = %ld", IoAdapter->Spid1[1]))
   DBG_REG(("  RbsConfig = %ld", IoAdapter->Spid1[2]))
   DBG_REG(("  RbsDigitTimeout = %ld", IoAdapter->Spid1[3]))
   DBG_REG(("  RbsBearerCap = %ld", IoAdapter->Spid1[4]))
   DBG_PRV0(("  RbsDebug = %x%x%x%x",
      IoAdapter->Spid1[5],IoAdapter->Spid1[6],IoAdapter->Spid1[7],IoAdapter->Spid1[8] ))
  }
 }
 if ( IoAdapter->L1TristateOrQsig )
 {
  DBG_REG(("  L1TristateOrQsig = 0x%x", IoAdapter->L1TristateOrQsig))
 }
 DBG_REG(("  DSPInfo = 0x%lx(0x%lx)",
   IoAdapter->InitialDspInfo, IoAdapter->InitialDspInfo << 14))
 a->ram_out (a, (void *) 8, (byte)IoAdapter->tei) ;
 a->ram_out (a, (void *) 9, (byte)IoAdapter->nt2) ;
 if ( IoAdapter->Properties.Family == FAMILY_MAESTRA )
  a->ram_out (a, (void *)10, (byte)IoAdapter->DidLen) ;
 else /* not availabble for S Cards */
  a->ram_out (a, (void *)10, (byte)0) ;
 a->ram_out (a, (void *)11, (byte)IoAdapter->WatchDog) ;
 a->ram_out (a, (void *)12, (byte)IoAdapter->Permanent) ;
 if ( IoAdapter->Properties.Family == FAMILY_MAESTRA )
  a->ram_out (a, (void *)13, (byte)IoAdapter->L1TristateOrQsig) ;
 else
  a->ram_out (a, (void *)13, (byte)0) ;
 a->ram_out (a, (void *)14, (byte)IoAdapter->StableL2) ;
 a->ram_out (a, (void *)15, (byte)IoAdapter->NoOrderCheck) ;
 if ( IoAdapter->Properties.Family == FAMILY_MAESTRA )
  a->ram_out (a, (void *)16, (byte)IoAdapter->ForceLaw) ;
 else
  a->ram_out (a, (void *)16, (byte)0) ;
 a->ram_out (a, (void *)17, (byte)0);
 a->ram_out (a, (void *)18, (byte)IoAdapter->LowChannel) ;
 a->ram_out (a, (void *)19, (byte)IoAdapter->ProtVersion) ;
 a->ram_out (a, (void *)20, (byte)IoAdapter->crc4) ;
 a->ram_out (a, (void *)21, (byte)(IoAdapter->NoHscx30 & 0x07));
 a->ram_out (a, (void *)22, (byte)IoAdapter->InitialDspInfo) ;
 if ( IoAdapter->Properties.Channels <= 2 )
 {
  CopyNumber (a, &IoAdapter->Oad1[0],   32) ;
  CopyNumber (a, &IoAdapter->Osa1[0],   64) ;
  CopyNumber (a, &IoAdapter->Spid1[0],  96) ;
  CopyNumber (a, &IoAdapter->Oad2[0],  128) ;
  CopyNumber (a, &IoAdapter->Osa2[0],  160) ;
  CopyNumber (a, &IoAdapter->Spid2[0], 192) ;
 } else if ( (IoAdapter->protocol_id == PROTTYPE_RBSCAS) && (i = IoAdapter->Spid1[0]) ) {
  /* take the Spid1 area, because Spids are not used */
  while ( i >= 0 ) {
   a->ram_out(a, (void *)(i+96), (byte)IoAdapter->Spid1[i]);
   i-- ;
  }
 }
/*
 * Special modem/fax tuneable parameters
 */
 if ( IoAdapter->ModemGuardTone )
 {
  DBG_REG(("  ModemGuardTone = %ld", IoAdapter->ModemGuardTone))
  offset = CopyUlongIfNonzero (a, PCINIT_MODEM_GUARD_TONE,
          IoAdapter->ModemGuardTone, 1, offset);
 }
 if  ( IoAdapter->ModemMinSpeed )
 {
  DBG_REG(("  ModemMinSpeed = %ld", IoAdapter->ModemMinSpeed))
  offset = CopyUlongIfNonzero (a, PCINIT_MODEM_MIN_SPEED,
          IoAdapter->ModemMinSpeed, 2, offset);
 }
 if  ( IoAdapter->ModemMaxSpeed )
 {
  DBG_REG(("  ModemMaxSpeed = %ld", IoAdapter->ModemMaxSpeed))
  offset = CopyUlongIfNonzero (a, PCINIT_MODEM_MAX_SPEED,
          IoAdapter->ModemMaxSpeed, 2, offset);
 }
 if  ( IoAdapter->ModemOptions )
 {
  DBG_REG(("  ModemOptions = 0x%08lx", IoAdapter->ModemOptions))
  offset = CopyUlongIfNonzero (a, PCINIT_MODEM_PROTOCOL_OPTIONS,
          IoAdapter->ModemOptions, 2, offset);
 }
 if  ( IoAdapter->ModemOptions2 )
 {
  DBG_REG(("  ModemOptions2 = 0x%08lx", IoAdapter->ModemOptions2))
  offset = CopyUlongIfNonzero (a, PCINIT_MODEM_OPTIONS,
          IoAdapter->ModemOptions2, 4, offset);
 }
 if  ( IoAdapter->ModemNegotiationMode )
 {
  DBG_REG(("  ModemNegotiationMode = 0x%08lx", IoAdapter->ModemNegotiationMode))
  offset = CopyUlongIfNonzero (a, PCINIT_MODEM_NEGOTIATION_MODE,
          IoAdapter->ModemNegotiationMode, 1, offset);
 }
 if  ( IoAdapter->ModemModulationsMask )
 {
  DBG_REG(("  ModemModulationsMask = 0x%08lx", IoAdapter->ModemModulationsMask))
  offset = CopyUlongIfNonzero (a, PCINIT_MODEM_MODULATIONS_MASK,
          IoAdapter->ModemModulationsMask, 3, offset);
 }
 if  ( IoAdapter->ModemTransmitLevel )
 {
  DBG_REG(("  ModemTransmitLevel = %ld", IoAdapter->ModemTransmitLevel))
  offset = CopyUlongIfNonzero (a, PCINIT_MODEM_TRANSMIT_LEVEL,
          IoAdapter->ModemTransmitLevel, 1, offset);
 }
 if  ( IoAdapter->ModemCarrierWaitTimeSec )
 {
  DBG_REG(("  Modem Carrier Wait Time = %d Sec",
        IoAdapter->ModemCarrierWaitTimeSec))
  offset = CopyUlongIfNonzero (a, PCINIT_MODEM_CARRIER_WAIT_TIME,
        (dword)IoAdapter->ModemCarrierWaitTimeSec, 1, offset);
 }
 if  ( IoAdapter->ModemCarrierLossWaitTimeTenthSec )
 {
  DBG_REG(("  Modem Carrier Loss Wait Time = %d Sec/10",
        IoAdapter->ModemCarrierLossWaitTimeTenthSec))
  offset = CopyUlongIfNonzero (a, PCINIT_MODEM_CARRIER_LOSS_TIME,
        (dword)IoAdapter->ModemCarrierLossWaitTimeTenthSec, 1, offset);
 }
 if  ( IoAdapter->FaxOptions )
 {
  DBG_REG(("  FaxOptions = 0x%08lx", IoAdapter->FaxOptions))
  offset = CopyUlongIfNonzero (a, PCINIT_FAX_OPTIONS,
          IoAdapter->FaxOptions, 2, offset);
 }
 if  ( IoAdapter->FaxMaxSpeed )
 {
  DBG_REG(("  FaxMaxSpeed = %ld", IoAdapter->FaxMaxSpeed))
  offset = CopyUlongIfNonzero (a, PCINIT_FAX_MAX_SPEED,
          IoAdapter->FaxMaxSpeed, 2, offset);
 }
 if  ( IoAdapter->Part68LevelLimiter )
 {
  DBG_REG(("  Part68LevelLimiter = %ld", IoAdapter->Part68LevelLimiter))
  offset = CopyUlongIfNonzero (a, PCINIT_PART68_LIMITER,
                               IoAdapter->Part68LevelLimiter, 1, offset);
 }
 if ( IoAdapter->QsigDialect )
 {
  DBG_REG(("  QsigDialect = 0x%x", IoAdapter->QsigDialect))
  offset = CopyUlongIfNonzero (a, PCINIT_QSIG_DIALECT,
                                 (dword)IoAdapter->QsigDialect, 1, offset);
 }
/*
 * Special North American parameters - BRI only
 */
 if ( IoAdapter->UsEktsNumCallApp ) {
  DBG_REG(("  Number Call Appearences %ld", IoAdapter->UsEktsNumCallApp))
  offset = CopyUlongIfNonzero (a, PCINIT_US_EKTS_CACH_HANDLES,
          IoAdapter->UsEktsNumCallApp, 1, offset);
 }
 if ( IoAdapter->UsEktsFeatAddConf ) {
  DBG_REG(("  Begin Conference Feature Activator = 0x%x", IoAdapter->UsEktsFeatAddConf))
  offset = CopyUlongIfNonzero (a, PCINIT_US_EKTS_BEGIN_CONF,
         (dword)IoAdapter->UsEktsFeatAddConf, 1, offset);
 }
 if ( IoAdapter->UsEktsFeatRemoveConf ) {
  DBG_REG(("  Drop Conference Feature Activator = =x%x", IoAdapter->UsEktsFeatRemoveConf))
  offset = CopyUlongIfNonzero (a, PCINIT_US_EKTS_DROP_CONF,
         (dword)IoAdapter->UsEktsFeatRemoveConf, 1, offset);
 }
 if ( IoAdapter->UsEktsFeatCallTransfer ) {
  DBG_REG(("  Call Transfer Feature Activator = 0x%x", IoAdapter->UsEktsFeatCallTransfer))
  offset = CopyUlongIfNonzero (a, PCINIT_US_EKTS_CALL_TRANSFER,
         (dword)IoAdapter->UsEktsFeatCallTransfer, 1, offset);
 }
 if ( IoAdapter->UsEktsFeatMsgWaiting ) {
  DBG_REG(("  Message Waiting = 0x%x", IoAdapter->UsEktsFeatMsgWaiting))
  offset = CopyUlongIfNonzero (a, PCINIT_US_EKTS_MWI,
         (dword)IoAdapter->UsEktsFeatMsgWaiting, 1, offset);
 }
 if ( IoAdapter->ForceVoiceMailAlert ) {
  DBG_REG(("  Force Voice Mail Alert = 0x%x", IoAdapter->ForceVoiceMailAlert))
  offset = CopyUlongIfNonzero (a, PCINIT_FORCE_VOICE_MAIL_ALERT,
         (dword)IoAdapter->ForceVoiceMailAlert, 1, offset);
 }
 if ( IoAdapter->DisableAutoSpid) {
  DBG_REG(("  Disable Auto SPID = 0x%x", IoAdapter->DisableAutoSpid))
  offset = CopyUlongIfNonzero (a, PCINIT_DISABLE_AUTOSPID_FLAG,
         (dword)IoAdapter->DisableAutoSpid, 1, offset);
 }
/*
 * it's possible to generate a ringtone, if the switch doesn't take this task
 */
 if ( IoAdapter->GenerateRingtone ) {
  DBG_REG(("  GenerateRingtone = 0x%x", IoAdapter->GenerateRingtone))
  offset = CopyUlongIfNonzero (a, PCINIT_RINGERTONE_OPTION,
         (dword)IoAdapter->GenerateRingtone, 1, offset);
 }
 if  ( IoAdapter->sdram_bar )
 {
  DBG_REG(("  CardAddress = 0x%08lx", IoAdapter->sdram_bar))
  offset = CopyUlongIfNonzero (a, PCINIT_CARD_ADDRESS,
          IoAdapter->sdram_bar, 4, offset);
 }
 if  ( IoAdapter->fpga_features )
 {
  DBG_REG(("  FPGAFeatures = 0x%08lx", IoAdapter->fpga_features))
  offset = CopyUlongIfNonzero (a, PCINIT_FPGA_FEATURES,
          IoAdapter->fpga_features, 1, offset);
 }
 xdiFeatures = 0 ;
 if  ( IoAdapter->features & PROTCAP_CMA_ALLPR )
  xdiFeatures |= PCINIT_XDI_CMA_FOR_ALL_NL_PRIMITIVES ;
 if ( xdiFeatures )
 {
  DBG_REG(("  XDIFeatures = 0x%08lx", xdiFeatures))
  offset = CopyUlongIfNonzero (a, PCINIT_XDI_FEATURES,
                               xdiFeatures, 1, offset);
 }
/*
 * Parameter for unchannelized protocols T1UNCH/E1UNCH
 */
 if ( IoAdapter->BChMask )
 {
  DBG_REG(("  BChMask = 0x%08lx", IoAdapter->BChMask))
  offset = CopyUlongIfNonzero (a, PCINIT_UNCHAN_B_MASK,
          IoAdapter->BChMask, 4, offset);
 }
/*
 * PIAFS link turnaround time, in PIAFS frames
 */
 if ( IoAdapter->PiafsLinkTurnaroundInFrames )
 {
  DBG_REG(("  PIAFS Link Turnaround = %d frames",
               IoAdapter->PiafsLinkTurnaroundInFrames))
  offset = CopyUlongIfNonzero (a, PCINIT_PIAFS_TURNAROUND_FRAMES,
               IoAdapter->PiafsLinkTurnaroundInFrames, 1, offset);
 }
/*
 * QSIG Features
 */
 if ( IoAdapter->QsigFeatures )
 {
  DBG_REG(("  QSIG Features = %04x", IoAdapter->QsigFeatures))
  offset = CopyUlongIfNonzero (a, PCINIT_QSIG_FEATURES,
               IoAdapter->QsigFeatures, 2, offset);
 }
/*
 * Parameter to prevent D channel signalling
 */ 
  if ( IoAdapter->nosig )
  {
    DBG_REG(("  nosig = %d", IoAdapter->nosig))
    offset = CopyUlongIfNonzero (a, PCINIT_NO_SIGNALLING,
                                IoAdapter->nosig, 1, offset);
  }
/*
 * Amount of TEI's that adapter will support in P2MP mode
 */ 
  if ( IoAdapter->BriLayer2LinkCount )
  {
    DBG_REG(("  L2 link count = %d", IoAdapter->BriLayer2LinkCount))
    offset = CopyUlongIfNonzero (a, PCINIT_L2_COUNT,
                                IoAdapter->BriLayer2LinkCount, 1, offset);
  }
/*
 * Adapter serial number
 */ 
  if ( IoAdapter->serialNo )
  {
    DBG_REG(("  Serial No = %d", IoAdapter->serialNo))
    offset = CopyUlongIfNonzero (a, PCINIT_CARD_SN,
                                IoAdapter->serialNo, 4, offset);
  }
/*
 * Adapter serial number
 */ 
  if ( IoAdapter->ControllerNumber )
  {
    DBG_REG(("  Port No = %d", IoAdapter->ControllerNumber))
    offset = CopyUlongIfNonzero (a, PCINIT_CARD_PORT,
                                (byte)IoAdapter->ControllerNumber, 1, offset);
  }
 a->ram_out (a, (void *)offset, 0);
}
