#include "platform.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>
#include "um_xdi.h"
#include "os.h"

#define DEBUG_OS
#undef DEBUG_OS 

extern void mlog_process_stdin_cmd (const char* cmd);

int diva_open_adapter (int adapter_nr) {
	int fd ;

	fd = open("/dev/DivasIDI", O_RDWR | O_NONBLOCK);
	if (fd < 0) {
#if defined(DEBUG_OS)
		printf ("A: can't open /dev/DivasIDI, errno=%d\n", errno);
#endif
		return (DIVA_INVALID_FILE_HANDLE);
	}
	if (write(fd, &adapter_nr, sizeof(adapter_nr)) != sizeof(adapter_nr)) {
#if defined(DEBUG_OS)
		printf ("A: can't set adapter number %d, errno=%d\n", adapter_nr, errno);
#endif
		close(fd);
		return (DIVA_INVALID_FILE_HANDLE);
	}
	return ((dword)fd);
}

int diva_close_adapter (int handle) {
	if (handle != DIVA_INVALID_FILE_HANDLE){
		close ((int)handle);
		return(0);
	}
#if defined(DEBUG_OS)
			printf ("A: close_adapter, errno=%d\n", errno);
#endif
	return(-1);
}

int diva_put_req (int handle, const void* data, int length) {
	int ret;
	
	if ((ret = write(handle, data, length)) < 0) {
#if defined(DEBUG_OS)
		printf ("A: put_req, errno=%d\n", errno);
#endif
	}

	return (ret);
}

int diva_get_message (int handle, void* data, int max_length) {
	int ret;
	
	if ((ret = read(handle, data, max_length)) < 0) {
#if defined(DEBUG_OS)
		printf ("A: get_message, errno=%d\n", errno);
#endif
	}

	return (ret);
}

int diva_wait_idi_message (int entity) {
  return (diva_wait_idi_message_interruptible (entity, 0));
}

int diva_wait_idi_message_interruptible (int entity, int interruptible) {
	fd_set readfds;
	struct timeval tv;
	int ret;

	for (;;) {
		memset (&readfds, 0x00, sizeof(readfds));

		FD_SET (entity, &readfds);
		tv.tv_sec = 0;
		tv.tv_usec = 100000;
		ret = select (entity+1, &readfds, 0, 0, &tv);
		if (ret < 0) {
			if ((errno == EINTR) && !interruptible)
				continue;
			else
				return (-1);
		}

		if (FD_ISSET(entity, &readfds)) {
			return (0);
		}
	}
}

unsigned long GetTickCount (void) {
	return (0);
}

void diva_wait_idi_message_and_stdin (int entity) {
	fd_set readfds;
	struct timeval tv;
	int ret;
	char tmp[256];
	int n;
	char* p;

	for (;;) {
		memset (&readfds, 0x00, sizeof(readfds));

		FD_SET (0,  &readfds);
		FD_SET (entity, &readfds);
		tv.tv_sec = 0;
		tv.tv_usec = 100000;
		ret = select (entity+1, &readfds, 0, 0, &tv);
		if (ret < 0) {
			return;
		}

		if (FD_ISSET(0, &readfds)) {
			n = read (0, tmp, sizeof(tmp)-1);
			tmp[n] = 0;
			if ((p = strstr (tmp, "\r"))) *p = 0;
			if ((p = strstr (tmp, "\n"))) *p = 0;
			mlog_process_stdin_cmd (&tmp[0]);
		}

		if (FD_ISSET(entity, &readfds)) {
			return ;
		}
	}
}
